/*
 * This file is part of the Owl Positioning System (OwlPS).
 * OwlPS is a project of the University of Franche-Comté
 * (Université de Franche-Comté), France.
 */


#ifndef _OWLPS_POSITIONING_RESULT_HH_
#define _OWLPS_POSITIONING_RESULT_HH_

class Request ;

#include "point3d.hh"

#include <string>

/// Represents a result computed by a positioning algorithm
class Result
{
protected:
  /// Request sent by the mobile
  const Request *request ;
  /// Algorithm used to compute the position
  std::string algorithm ;
  /// Computed coordinates of the mobile
  Point3D position ;
  /// Distance error between the real coordinates and the computed point
  /** The error is set to -1 if the real coordinates are unknown. */
  float error ;

public:
  Result(const Request *_request = NULL,
         const std::string &_algorithm = "UnknownAlgorithm"):
    request(_request), algorithm(_algorithm), error(-1) {}
  Result(const Request *_request, const std::string &_algorithm,
         const Point3D &_position):
    request(_request), algorithm(_algorithm), position(_position),
    error(-1) {}
  Result(const Request *_request, const std::string &_algorithm,
         const Point3D &_position, const Point3D &real_position) ;
  ~Result(void) {}

  /** @name Read accessors */
  //@{
  const Request* get_request(void) const ;
  const std::string& get_algorithm(void) const ;
  const Point3D& get_position(void) const ;
  float get_error(void) const ;
  std::string in_which_area(void) const ;
  //@}

  /** @name Write accessors */
  //@{
  void compute_error(const Point3D &real_position) ;
  //@}

  /** @name Operators */
  //@{
  Result& operator=(const Result &source) ;
  bool operator==(const Result &source) const ;
  bool operator!=(const Result &source) const ;
  //@}

  /** @name Conversion accessors */
  //@{
  /// Converts to a CSV string
  const std::string to_csv(void) const ;
  //@}

  /// Displays a Result
  friend std::ostream& operator<<(std::ostream &os, const Result &r) ;
} ;



/* *** Read accessors *** */


inline const Request* Result::get_request() const
{
  return request ;
}


inline const std::string& Result::get_algorithm() const
{
  return algorithm ;
}


inline const Point3D& Result::get_position() const
{
  return position ;
}


inline float Result::get_error() const
{
  return error ;
}



/* *** Operators *** */


inline bool Result::operator!=(const Result &source) const
{
  return !(*this == source) ;
}



#endif // _OWLPS_POSITIONING_RESULT_HH_
